 //////////////////////////////////////////////////////////////////////////////
//
// Syntax Highlighter, v1.0. Copyright  1998 Midnight Beach, all rights
// reserved. Reads C++ and Pascal source from the clipboard, writes it back
// syntax highlighted in HTML and RTF. Feel free to use the executable however
// you like, without any royalties or licensing. Please do NOT redistribute
// the executable, though - I would like the sole source to stay
// "http://www.midnightbeach.com/ftp_page/SyntaxHighlighter.html". Similarly,
// feel free to modify the source as you wish, but please do not redistribute
// the original source or any changes which you make. Questions and/or comments
// to jon@midnightbeach.com
//

//---------------------------------------------------------------------------

#define UseVector 0
#define UseSet    1
#define Benchmark 0

#include <vcl.h>
#pragma hdrstop

#include <assert.h>
#include "Convert.h"

#if UseVector
#include <vector>
#else
#include <list>
#endif

#if UseSet
#include <set>
#endif

using std::for_each;

//
// Abstract tokenizer
//

struct Token {
       TokenTypes Type;
       string     Value;

       Token() { Type = Unknown; Value = ""; };
       Token(TokenTypes T, string& V) : Type(T), Value(V) {};
}; // Token

#if UseVector
typedef std::vector<Token> TokenStream;
#else
typedef std::list<Token> TokenStream;
#endif

typedef string KeywordType;
#if UseSet
typedef std::set<KeywordType> KeywordSet;
#endif

class Parser 
{
  private:
    TokenStream* Result;
    #if UseSet
    KeywordSet* Keywords;
    #else
    KeywordType* begin_Keyword;
    KeywordType* end_Keyword;
    #endif

    void DoNormal(   char C, string& StringC);

  protected:
    string       TokenText;
    TokenTypes   TokenType;
    string       LookFor;
    unsigned int LookCount;
    bool         CaseSensitive;

    #if UseSet
    Parser(bool Case, KeywordSet* _Keywords);
    #else
    Parser(bool Case, KeywordType* begin, KeywordType* end);
    #endif

    typedef void (Parser::* TypeHandler)(char C, string& StringC);
//    typedef void (__closure* TypeHandler)(char C, string& StringC);
    TypeHandler Handler[Unknown+1];

    void Set(TokenTypes Type, string& Text)
    { TokenType = Type; TokenText = Text; };
    void SetV(TokenTypes Type, string Text) { Set(Type, Text); };
    void Reset() { SetV(Unknown, ""); TokenText.reserve(24); };
    void Output(TokenTypes TokenType, string& TokenText) const
    { Result->push_back(Token(TokenType, TokenText)); }
    void OutputV(TokenTypes TokenType, string TokenText) const
    { Output(TokenType, TokenText); }
    void OutputAndReset(TokenTypes TokenType, string& TokenText)
    { Output(TokenType, TokenText); Reset(); }

  public:
    void SetTokenStream(TokenStream& ResultList)
    { Result = & ResultList; Reset(); };
    inline void operator()(char C);
    ~Parser();
}; // Parser

#if UseSet
Parser::Parser(bool Case, KeywordSet* _Keywords)
: CaseSensitive(Case), Keywords(_Keywords)
#else
Parser::Parser(bool Case, KeywordType* begin, KeywordType* end)
: CaseSensitive(Case), begin_Keyword(begin), end_Keyword(end)
#endif
{
  Handler[Normal] = & Parser::DoNormal;
} // Parser::Parser

void Parser::operator()(char C)
{
  assert(Handler[TokenType] != NULL);
  string StringC = string(1, C);
  (this->*Handler[TokenType])(C, StringC);
} // Parser::operator()

Parser::~Parser()
{
  if (TokenText != "") Output( (TokenType != Unknown) ? TokenType : Normal,
                               TokenText );
} // Parser::~Parser

bool IsAlpha(char C)        { return isalpha(C) || C == '_'; }
bool IsAlphaNumeric(char C) { return isalnum(C) || C == '_'; }

void Parser::DoNormal(char C, string& StringC)
{
  if (IsAlphaNumeric(C)) TokenText += StringC;
  else {
    string Token = TokenText;
    if (! CaseSensitive)
      std::transform(Token.begin(), Token.end(), Token.begin(), tolower);
    #if UseSet
    bool IsKeyword = Keywords->count(Token) != 0;
    #else
    using std::find_if;
    using std::bind2nd;
    using std::equal_to;

    KeywordType* KPtr = find_if( begin_Keyword, end_Keyword,
                                 bind2nd(equal_to<string>(), Token) );
    bool IsKeyword = KPtr != end_Keyword;
    #endif
    OutputAndReset( (IsKeyword) ? Keyword : Normal,
                    TokenText );
    TokenText = StringC;
  }
} // Parser::DoNormal

class ParserWrapper
{
  private:
    Parser* P;

  public:
    ParserWrapper(Parser& _P)  { P = & _P; };
    void operator()(char C) const { P->operator()(C); };
};

void Tokenize(TokenStream& Result, SourceData Source, Parser& P)
{
  P.SetTokenStream(Result);
  for_each(Source, Source + strlen(Source), ParserWrapper(P));
}

// Specific parsers

#define XSet(Type, StringC) { \
  if (TokenText != "") Output(Normal, TokenText); \
  Set(Type, StringC); }

#include "CppParser.cpp"
#include "DelphiParser.cpp"

#undef XSet

//
// Output generator
//

#if UseVector
typedef std::vector<string> OutputList;
#else
typedef std::list<string> OutputList;
#endif

HGLOBAL ConsolidateOutputList(string Prefix, OutputList& List, string Suffix)
{
  struct GetLengths {
    int* Length;
    GetLengths(int& L) : Length(& L) {};
    inline operator () (string S) const { *Length += S.length(); }
  }; // GetLengths

  struct Append {
    char* End;
    Append(char* E) : End(E) {};
    inline operator () (string S) { int Length = S.length();
                                    memcpy(End, S.c_str(), Length);
                                    End += Length; }
  }; // Append

  int PrefixLength = Prefix.length(),
      SuffixLength = Suffix.length(),
      Length = PrefixLength + SuffixLength;
  for_each(List.begin(), List.end(), GetLengths(Length));

  HGLOBAL Result = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, Length + 1);
  char* Buffer = (char*) GlobalLock(Result);
  memcpy(Buffer, Prefix.c_str(), PrefixLength); // no \0 at end!
  for_each(List.begin(), List.end(), Append(Buffer + PrefixLength));
  strcpy(Buffer + Length - SuffixLength, Suffix.c_str());
  GlobalUnlock(Result);

  return Result;
} // ConsolidateOutputList

class TokenToOutput
{
  private:
    OutputList* Result;

  protected:
    inline void Output(string& S) { Result->push_back(S); };
    inline void OutputV(string S) { Result->push_back(S); };
    inline virtual void SetFormat(Token& T) = 0;
    inline virtual void ResetFormat(Token& T) = 0;
    inline virtual string DoQuoting(string& S) = 0;

  public:
    TokenToOutput(OutputList& R) : Result(& R) {};
    inline void operator () (Token& T) ;
}; // TokenToOutput

void TokenToOutput::operator () (Token& T)
{
  SetFormat(T);
  Output( DoQuoting(T.Value) );
  ResetFormat(T);
} // TokenToOutput::operator ()

// specific formatters

#include "TokenToHtml.cpp"
#include "TokenToRtf.cpp"

//
// Entry points
//

// Immediate rendering

HGLOBAL CppToHtml(SourceData Source)
{
  TokenStream Tokens;
  Tokenize(Tokens, Source, CppParser());
  return TokensToHtml(Tokens);
} // CppToHtml

HGLOBAL DelphiToHtml(SourceData Source)
{
  TokenStream Tokens;
  Tokenize(Tokens, Source, DelphiParser());
  return TokensToHtml(Tokens);
} // DelphiToHtml

HGLOBAL CppToRtf(SourceData Source)
{
  TokenStream Tokens;
  Tokenize(Tokens, Source, CppParser());
  return TokensToRtf(Tokens);
} // CppToRtf

HGLOBAL DelphiToRtf(SourceData Source)
{
  TokenStream Tokens;
  Tokenize(Tokens, Source, DelphiParser());
  return TokensToRtf(Tokens);
} // DelphiToRtf

HGLOBAL Highlight( OutputFormats Output,
                  SourceData Source,
                  InputLanguages Input)
{
  if (Output == HTML)
    if (Input == Cpp) return CppToHtml(Source);
    else return DelphiToHtml(Source);
  else if (Input == Cpp) return CppToRtf(Source);
       else return DelphiToRtf(Source);
} // Highlight

#if Benchmark

int msecs(TDateTime Delta)
{
  return (double) Delta * 24 * 60 * 60 * 1000;
}

void BothFormats( HGLOBAL& HTML, HGLOBAL& RTF,
                  SourceData Source, InputLanguages Language)
{
TDateTime T1, T2, T3, T4;

T1 = Now();

  TokenStream Tokens;
  if (Language == Cpp) Tokenize(Tokens, Source, CppParser());
  else                 Tokenize(Tokens, Source, DelphiParser() ) ;

T2 = Now();
  HTML = TokensToHtml(Tokens);
T3 = Now();
  RTF  = TokensToRtf(Tokens);
T4 = Now();
ShowMessage( (AnsiString)
             ("Tokenizing = " + IntToString(msecs(T2 - T1)) + "\r\n"
              "Htmlizing  = " + IntToString(msecs(T3 - T2)) + "\r\n"
              "Rtfizing   = " + IntToString(msecs(T4 - T3)) + "\r\n").c_str() );
} // BothFormats

#else

void BothFormats( HGLOBAL& HTML, HGLOBAL& RTF,
                  SourceData Source, InputLanguages Language)
{
  TokenStream Tokens;
  if (Language == Cpp) Tokenize(Tokens, Source, CppParser());
  else                 Tokenize(Tokens, Source, DelphiParser() ) ;

  HTML = TokensToHtml(Tokens);
  RTF  = TokensToRtf(Tokens);
} // BothFormats

#endif

// Delayed rendering

TokenStream Tokens;

void DiscardParseData()
{
  Tokens.resize(0);
} // DiscardParseData

#if Benchmark

void Parse(SourceData Source, InputLanguages Language)
{
  DiscardParseData();

TDateTime T1, T2;
T1 = Now();
  if (Language == Cpp) Tokenize(Tokens, Source, CppParser() );
  else                 Tokenize(Tokens, Source, DelphiParser() ) ;
T2 = Now();
ShowMessage( (AnsiString)
             ("Tokenizing = " + IntToString(msecs(T2 - T1))).c_str() );

} // Parse

#else

void Parse(SourceData Source, InputLanguages Language)
{
  DiscardParseData();
  if (Language == Cpp) Tokenize(Tokens, Source, CppParser() );
  else                 Tokenize(Tokens, Source, DelphiParser() ) ;
} // Parse

#endif

HGLOBAL RenderToHtml()
{
  return TokensToHtml(Tokens);
} // RenderToHtml

HGLOBAL RenderToRTF()
{
  return TokensToRtf(Tokens);
} // RenderToRTF

//---------------------------------------------------------------------------
#pragma package(smart_init)

